
;(format *terminal-io* "~%; Loading /lisp/pcl init file. ~%")

(excl:set-case-mode :case-insensitive-upper)
(setq excl:*ignore-package-name-case* t)

;(excl::generate-library-pathnames "/software/cl/lib")




;;; add these function to your own lisp-init file to auto-load the
;;; belief packages when you call (load-bel <filename>)

;;; This contains the home of the BELIEF package code
(defvar bel-home (make-pathname ; :host "hustat" :device "xy0d"
				:directory "~almond/belief1.1/lisp"
				;:type "lisp"
				))
;;; note change type to "lisp" or "bin" to use uncompiled or compiled
;;; under default conditions files respectively.  

(require 'belief-init (merge-pathnames "belief-init" bel-home))

;;; Set up [] and {} as matching parenthesis (for constructing ps-sets
;;; and other nested lists.


;; read-match, reads to the next matching closing parenthesis ),], or}
;; respectively.  
(defun read-match (stream char)
    (declare (character char))
    (read-delimited-list (case char (#\( #\)) (#\[ #\]) (#\{ #\}))
			 stream t))

;; set up {} and [] as macro characters
(set-macro-character #\{ #'read-match nil)
(set-macro-character #\} (get-macro-character #\)))
(set-macro-character #\[ #'read-match nil)
(set-macro-character #\] (get-macro-character #\)))

(if (equal "Allegro CL" (lisp-implementation-type))
    (tpl:setq-default lisp:*readtable* lisp:*readtable*) ; necessary in
					; Allegro CL because of
					; readtable problems, now
					; default should be set
					; properly. 
)


(defvar *belief-format* t		;defaults to belief format
  "Should belief functions (t) or potentials (nil) be used as a default")




;;; load-belief-package
;;; auto-loading function loads various parts of the belief package.
;;; If neither :belief nor :prob is specified, :belief is loaded by
;;; default.  The function :belq causes the user package to use the
;;; belief package, and may cause strange results when switching back
;;; and forth between belief and prob modes.
;; This should be wrapped in a (eval-when (compile load eval))
(defun load-belief-package (&key belief prob belq cost search nuke)
  (if (and (null belief) (null prob)) (setq belief t))
  (when prob (require 'probread (merge-pathnames "probread" bel-home))
	(when nuke (require 'distprob (merge-pathnames "distprob" bel-home)))
	(setq *belief-format* nil))
  (when belief (require 'read (merge-pathnames "read" bel-home))
	(when nuke (require 'distbel (merge-pathnames "distbel" bel-home)))
	(setq *belief-format* t))
  (require 'computations (merge-pathnames "computations" bel-home))
  (when belq (require 'belp (merge-pathnames "belq" bel-home))
	(mode:bel-mode 'user))
  (if search (require 'search (merge-pathnames "search" bel-home)))
  (if cost (require 'cost (merge-pathnames "cost" bel-home)))
  (when nuke (require 'readnuke (merge-pathnames "readnuke" bel-home))
	(require 'monte (merge-pathnames "monte" bel-home)))
  (use-package '(shell basic graphs inherited subshell)))

;;; Theses constants are used by the basic package to initialize the
;;; size of the hash table
;;; Global constants -- later versions should query for these
(in-package 'basic)
(defparameter *maxvars* 5 "Maximum number of variables to combine over")
(defparameter *maxvals* 2 "Maximum number of values taken on by one variable")
(in-package 'user)

;;; load-functions for data bases.

;; load-bel load a belief function data base (in belief mode)
(defun load-bel (filename &key nuke)
  ;(franz-kludge)
  (load-belief-package :belief t :nuke nuke)
  (load filename)
  (princ "Functions: do-it redo-it undo-it ppnode ppmessages ppval ppall")
  (terpri))


;; load-prob load data base in prob mode.
(defun load-prob (filename &key nuke)
  ;(franz-kludge)
  (load-belief-package :belief nil :prob t :nuke nuke)
  (load filename)
  (princ "Functions: do-it redo-it undo-it ppnode ppmessages ppval ppall")
  (terpri))


(setq *read-default-float-format* 'long-float)


;;; Random state setting routine.  This "randomly" initializes the
;;; random state. 
;;; See Steele, Section 12.9 for information on how to use a fixed
;;; seed
(setq *random-state* (make-random-state t))

;;; kludge for read-macros in Franz (Extended) Common Lisp

(if  nil ;(equal "Allegro CL" (lisp-implementation-type))
(defun franz-kludge ()
	(set-macro-character #\{ #'read-match nil)
	(set-macro-character #\} (get-macro-character #\)))
	(set-macro-character #\[ #'read-match nil)
	(set-macro-character #\] (get-macro-character #\)))
	(set-dispatch-macro-character #\# #\? #'will:|#?-reader|)
	(set-dispatch-macro-character #\# #\! #'will:|#!-reader|))
;else (ordinary lisps)
(defun franz-kludge () ))




